@@ -5,23 +5,30 @@ import numpy as np
 
 from phi import struct
 from phi.backend.dynamic_backend import DYNAMIC_BACKEND as math
-from .helper import _get_pad_width_axes, _get_pad_width, spatial_rank, _dim_shifted, _contains_axis, spatial_dimensions, all_dimensions
+from phi.struct.functions import mappable
+
+from .helper import (_contains_axis, _dim_shifted, _get_pad_width,
+                     _get_pad_width_axes, all_dimensions, rank,
+                     spatial_dimensions, spatial_rank)
 
 
-def indices_tensor(tensor, dtype=np.float32):
+def indices_tensor(tensor, dtype=None):
     """
     Returns an index tensor of the same spatial shape as the given tensor.
     Each index denotes the location within the tensor starting from zero.
     Indices are encoded as vectors in the index tensor.
 
     :param tensor: a tensor of shape (batch size, spatial dimensions..., component size)
-    :param dtype: a numpy data type (default float32)
+    :param dtype: NumPy data type or `None` for default
     :return: an index tensor of shape (1, spatial dimensions..., spatial rank)
     """
     spatial_dimensions = list(tensor.shape[1:-1])
     idx_zyx = np.meshgrid(*[range(dim) for dim in spatial_dimensions], indexing='ij')
     idx = np.stack(idx_zyx, axis=-1).reshape([1, ] + spatial_dimensions + [len(spatial_dimensions)])
-    return idx.astype(dtype)
+    if dtype is not None:
+        return idx.astype(dtype)
+    else:
+        return math.to_float(idx)
 
 
 def normalize_to(target, source=1, epsilon=1e-5, batch_dims=1):
@@ -59,7 +66,10 @@ def batch_align(tensor, innate_dims, target, convert_to_same_backend=True):
 
 
 def batch_align_scalar(tensor, innate_spatial_dims, target):
-    if math.staticshape(tensor)[-1] != 1:
+    if rank(tensor) == 0:
+        assert innate_spatial_dims == 0
+        return math.expand_dims(tensor, 0, len(math.staticshape(target)))
+    if math.staticshape(tensor)[-1] != 1 or math.ndims(tensor) <= 1:
         tensor = math.expand_dims(tensor, -1)
     result = batch_align(tensor, innate_spatial_dims + 1, target)
     return result
@@ -80,9 +90,9 @@ Runs a blur kernel over the given tensor.
         cutoff = min(int(round(radius * 3)), *field.shape[1:-1])
 
     xyz = np.meshgrid(*[range(-int(cutoff), (cutoff) + 1) for _ in field.shape[1:-1]])
-    d = np.float32(np.sqrt(np.sum([x**2 for x in xyz], axis=0)))
+    d = math.to_float(np.sqrt(np.sum([x**2 for x in xyz], axis=0)))
     if kernel == "1/1+x":
-        weights = np.float32(1) / (d / radius + 1)
+        weights = math.to_float(1) / (d / radius + 1)
     elif kernel.lower() == "gauss":
         weights = math.exp(- d / radius / 2)
     else:
@@ -123,6 +133,29 @@ def l_n_loss(tensor, n, batch_norm=True):
         return total_loss
 
 
+def frequency_loss(tensor, frequency_falloff=100, reduce_batches=True):
+    """
+    Instead of minimizing each entry of the tensor, minimize the frequencies of the tensor, emphasizing lower frequencies over higher ones.
+
+    :param reduce_batches: whether to reduce the batch dimension of the loss by adding the losses along the first dimension
+    :param tensor: typically actual - target
+    :param frequency_falloff: large values put more emphasis on lower frequencies, 1.0 weights all frequencies equally.
+    :return: scalar loss value
+    """
+    if struct.isstruct(tensor):
+        all_tensors = struct.flatten(tensor)
+        return sum(frequency_loss(tensor, frequency_falloff, reduce_batches) for tensor in all_tensors)
+    diff_fft = abs_square(math.fft(tensor))
+    k = fftfreq(tensor.shape[1:-1], mode='absolute')
+    weights = math.exp(-0.5 * k ** 2 * frequency_falloff ** 2)
+    return l1_loss(diff_fft * weights, reduce_batches=reduce_batches)
+
+
+@mappable()
+def abs_square(complex):
+    return math.imag(complex) ** 2 + math.real(complex) ** 2
+
+
 # Divergence
 
 def divergence(tensor, dx=1, difference='central'):
@@ -200,11 +233,13 @@ def axis_gradient(tensor, spatial_axis):
 
 # Laplace
 
-def laplace(tensor, padding='replicate', axes=None):
+@mappable()
+def laplace(tensor, padding='replicate', axes=None, use_fft_for_periodic=False):
     """
     Spatial Laplace operator as defined for scalar fields.
     If a vector field is passed, the laplace is computed component-wise.
 
+    :param use_fft_for_periodic: If True and padding='circular', uses FFT to compute laplace
     :param tensor: n-dimensional field of shape (batch, spacial dimensions..., components)
     :param padding: 'valid', 'constant', 'reflect', 'replicate', 'circular'
     :param axes: The second derivative along these axes is summed over
@@ -214,7 +249,7 @@ def laplace(tensor, padding='replicate', axes=None):
     rank = spatial_rank(tensor)
     if padding is None or padding == 'valid':
         pass  # do not pad tensor
-    elif padding in ['circular', 'wrap']:
+    elif padding in ('circular', 'wrap') and use_fft_for_periodic:
         return fourier_laplace(tensor)
     else:
         tensor = math.pad(tensor, _get_pad_width_axes(rank, axes, val_true=[1, 1], val_false=[0, 0]), padding)
@@ -230,7 +265,7 @@ def laplace(tensor, padding='replicate', axes=None):
 
 
 def _conv_laplace_2d(tensor):
-    kernel = np.array([[0., 1., 0.], [1., -4., 1.], [0., 1., 0.]], dtype=np.float32)
+    kernel = math.to_float([[0., 1., 0.], [1., -4., 1.], [0., 1., 0.]])
     kernel = kernel.reshape((3, 3, 1, 1))
     if tensor.shape[-1] == 1:
         return math.conv(tensor, kernel, padding='VALID')
@@ -254,10 +289,9 @@ def _conv_laplace_3d(tensor):
 
     padding explicitly done in laplace(), hence here not needed
     """
-    kernel = np.array([[[0., 0., 0.], [0., 1., 0.], [0., 0., 0.]],
-                       [[0., 1., 0.], [1., -6., 1.], [0., 1., 0.]],
-                       [[0., 0., 0.], [0., 1., 0.], [0., 0., 0.]]],
-                      dtype=np.float32)
+    kernel = math.to_float([[[0., 0., 0.], [0., 1., 0.], [0., 0., 0.]],
+                            [[0., 1., 0.], [1., -6., 1.], [0., 1., 0.]],
+                            [[0., 0., 0.], [0., 1., 0.], [0., 0., 0.]]])
     kernel = kernel.reshape((3, 3, 3, 1, 1))
     if tensor.shape[-1] == 1:
         return math.conv(tensor, kernel, padding='VALID')
@@ -281,18 +315,52 @@ def _sliced_laplace_nd(tensor, axes=None):
     return math.sum(components, 0)
 
 
-def fourier_laplace(tensor):
+@mappable()
+def fourier_laplace(tensor, times=1):
+    """
+Applies the spatial laplce operator to the given tensor with periodic boundary conditions.
+
+*Note:* The results of `fourier_laplace` and `laplace` are close but not identical.
+
+This implementation computes the laplace operator in Fourier space.
+The result for periodic fields is exact, i.e. no numerical instabilities can occur, even for higher-order derivatives.
+    :param tensor: tensor, assumed to have periodic boundary conditions
+    :param times: number of times the laplace operator is applied. The computational cost is independent of this parameter.
+    :return: tensor of same shape as `tensor`
+    """
     frequencies = math.fft(math.to_complex(tensor))
     k = fftfreq(math.staticshape(tensor)[1:-1], mode='square')
     fft_laplace = -(2 * np.pi)**2 * k
-    return math.real(math.ifft(frequencies * fft_laplace))
+    return math.real(math.ifft(frequencies * fft_laplace ** times))
 
 
-def fftfreq(resolution, mode='vector', dtype=np.float32):
+@mappable()
+def fourier_poisson(tensor, times=1):
+    """ Inverse operation to `fourier_laplace`. """
+    frequencies = math.fft(math.to_complex(tensor))
+    k = fftfreq(math.staticshape(tensor)[1:-1], mode='square')
+    fft_laplace = -(2 * np.pi)**2 * k
+    fft_laplace[(0,) * math.ndims(k)] = np.inf
+    return math.cast(math.real(math.ifft(math.divide_no_nan(frequencies, fft_laplace**times))), math.dtype(tensor))
+
+
+def fftfreq(resolution, mode='vector', dtype=None):
+    """
+    Returns the discrete Fourier transform sample frequencies.
+    These are the frequencies corresponding to the components of the result of `math.fft` on a tensor of shape `resolution`.
+
+    :param resolution: grid resolution measured in cells
+    :param mode: one of (None, 'vector', 'absolute', 'square')
+    :param dtype: data type of the returned tensor
+    :return: tensor holding the frequencies of the corresponding values computed by math.fft
+    """
     assert mode in ('vector', 'absolute', 'square')
     k = np.meshgrid(*[np.fft.fftfreq(int(n)) for n in resolution], indexing='ij')
     k = math.expand_dims(math.stack(k, -1), 0)
-    k = k.astype(dtype)
+    if dtype is not None:
+        k = k.astype(dtype)
+    else:
+        k = math.to_float(k)
     if mode == 'vector':
         return k
     k = math.sum(k**2, axis=-1, keepdims=True)
@@ -304,21 +372,23 @@ def fftfreq(resolution, mode='vector', dtype=np.float32):
 
 # Downsample / Upsample
 
-def downsample2x(tensor, interpolation='linear'):
+def downsample2x(tensor, interpolation='linear', axes=None):
     if struct.isstruct(tensor):
-        return struct.map(lambda s: downsample2x(s, interpolation),
+        return struct.map(lambda s: downsample2x(s, interpolation, axes),
                           tensor, recursive=False)
 
     if interpolation.lower() != 'linear':
         raise ValueError('Only linear interpolation supported')
-    dims = range(spatial_rank(tensor))
+    rank = spatial_rank(tensor)
+    if axes is None:
+        axes = range(rank)
     tensor = math.pad(tensor,
                       [[0, 0]]
-                      + [([0, 1] if (dim % 2) != 0 else [0, 0]) for dim in tensor.shape[1:-1]]
+                      + [([0, 1] if (dim % 2) != 0 and _contains_axis(axes, ax, rank) else [0, 0]) for ax, dim in enumerate(tensor.shape[1:-1])]
                       + [[0, 0]], 'replicate')
-    for dimension in dims:
-        upper_slices = tuple([(slice(1, None, 2) if i == dimension else slice(None)) for i in dims])
-        lower_slices = tuple([(slice(0, None, 2) if i == dimension else slice(None)) for i in dims])
+    for axis in axes:
+        upper_slices = tuple([(slice(1, None, 2) if i == axis else slice(None)) for i in range(rank)])
+        lower_slices = tuple([(slice(0, None, 2) if i == axis else slice(None)) for i in range(rank)])
         tensor_sum = tensor[(slice(None),) + upper_slices + (slice(None),)] + tensor[(slice(None),) + lower_slices + (slice(None),)]
         tensor = tensor_sum / 2
     return tensor
