@@ -1,10 +1,7 @@
 import warnings
-
 import six
-from phi.struct.context import skip_validate
-from phi.struct.structdef import Item
 
-from .physics import Physics, State, struct
+from .physics import Physics, State, struct, _ChainedPhysics, _as_physics
 
 
 class StateCollection(dict):
@@ -191,7 +188,7 @@ class CollectivePhysics(Physics):
         return self.physics[state.name] if state.name in self.physics else state.default_physics()
 
     def add(self, name, physics):
-        self.physics[name] = physics
+        self.physics[name] = _as_physics(physics)
 
     def remove(self, name):
         if name in self.physics:
