@@ -1,8 +1,8 @@
 from .context import unsafe
-from .item_condition import DATA, VARIABLES, CONSTANTS, ALL_ITEMS
 from .trait import Trait
 from .structdef import definition, variable, constant, derived
+from .item_condition import DATA, VARIABLES, CONSTANTS, ALL_ITEMS, ignore
 from .struct import Struct, kwargs, to_dict, variables, constants, properties_dict, copy_with, isstruct, equal, VALID, INVALID
 
 # pylint: disable-msg = redefined-builtin
-from .functions import flatten, names, map, map_item, zip, Trace, compare, print_differences, shape, staticshape, dtype
+from .functions import flatten, unflatten, names, map, map_item, zip, Trace, compare, print_differences, shape, staticshape, dtype, any, all
