@@ -1,25 +1,31 @@
-import warnings
-
 import numpy as np
 import phi.app.app as base_app
 import six
-import tensorflow as tf
-from phi.app.app import EditableFloat, EditableInt, EditableValue
-from phi.data.reader import BatchReader
 
+from . import tf
+from phi.app.app import EditableFloat, EditableInt, EditableValue
+from phi.data.dataset import Dataset
+from phi.data.reader import BatchReader
+from phi.physics.field import Field, StaggeredGrid
+from phi.tf.data import Dataset as TFDataset
+
+from . import TF_BACKEND
 from .session import Session
-from .util import istensor
 from .world import tf_bake_graph
+from .. import math
+from ..app import EditableBool
 
 
 class App(base_app.App):
 
     def __init__(self, *args, **kwargs):
         base_app.App.__init__(self, *args, **kwargs)
-        self.session = Session(self.scene)
+        config = tf.ConfigProto()
+        config.gpu_options.allow_growth = True
+        self.session = Session(self.scene, session=tf.Session(config=config))
         self.scalars = []
         self.scalar_names = []
-        self.editable_placeholders = {}
+        self.editable_placeholders = {}  # placeholder -> attribute name
         self.auto_bake = True
         self.add_trait('tensorflow')
 
@@ -27,41 +33,55 @@ class App(base_app.App):
         if self.prepared:
             return
         base_app.App.prepare(self)
+        if self.auto_bake:
+            self.info("Baking static TensorFlow graph (disable by setting 'app.auto_bake = False'")
+            tf_bake_graph(self.world, self.session)
         self.info('Initializing variables')
         self.session.initialize_variables()
-        if self.auto_bake:
-            tf_bake_graph(self.world, self.session)
         return self
 
     def add_scalar(self, name, node):
-        assert isinstance(node, tf.Tensor)
+        assert TF_BACKEND.is_tensor(node), 'add_scalar requires a TensorFlow tensor but got %s' % node
         self.scalar_names.append(name)
         self.scalars.append(node)
 
     def editable_float(self, name, initial_value, minmax=None, log_scale=None):
         val = EditableFloat(name, initial_value, minmax, None, log_scale)
-        setattr(self, 'float_' + name.lower(), val)
-        placeholder = tf.placeholder(tf.float32, (), name.lower().replace(' ', '_'))
+        self.set_editable_value(name, val)
+        placeholder = tf.placeholder(TF_BACKEND.precision_dtype, (), name.lower().replace(' ', '_'))
         self.add_scalar(name, placeholder)
-        self.editable_placeholders[placeholder] = 'float_' + name.lower()
+        self.editable_placeholders[placeholder] = name
         return placeholder
 
     def editable_int(self, name, initial_value, minmax=None):
         val = EditableInt(name, initial_value, minmax, None)
-        setattr(self, 'int_' + name.lower(), val)
+        self.set_editable_value(name, val)
         placeholder = tf.placeholder(tf.int32, (), name.lower().replace(' ', '_'))
         self.add_scalar(name, placeholder)
-        self.editable_placeholders[placeholder] = 'int_' + name.lower()
+        self.editable_placeholders[placeholder] = name
+        return placeholder
+
+    def editable_bool(self, name, initial_value, add_as_scalar=False):
+        val = EditableBool(name, initial_value)
+        self.set_editable_value(name, val)
+        placeholder = tf.placeholder(tf.bool, (), name.lower().replace(' ', '_'))
+        if add_as_scalar:
+            self.add_scalar(name, math.to_float(placeholder))
+        self.editable_placeholders[placeholder] = name
         return placeholder
 
     def editable_values_dict(self):
-        feed_dict = {}
-        for placeholder, attrname in self.editable_placeholders.items():
-            val = getattr(self, attrname)
-            if isinstance(val, EditableValue):
-                val = val.initial_value
-            feed_dict[placeholder] = val
-        return feed_dict
+        return {placeholder: self.get_editable_value(name) for placeholder, name in self.editable_placeholders.items()}
+
+    def get_editable_value(self, name):
+        value = getattr(self, '_ed_val_' + name.lower())
+        if isinstance(value, EditableValue):
+            return value.initial_value
+        else:
+            return value
+
+    def set_editable_value(self, name, value):
+        setattr(self, '_ed_val_' + name.lower(), value)
 
 
 def EVERY_EPOCH(tfapp): return tfapp.steps % tfapp.epoch_size == 0
@@ -82,15 +102,22 @@ class LearningApp(App):
                  **kwargs):
         App.__init__(self, name=name, subtitle=subtitle, base_dir=base_dir, **kwargs)
         self.add_trait('model')
-        self.learning_rate = self.editable_float('Learning_Rate', learning_rate)
-        self.training = tf.placeholder(tf.bool, (), 'training')
-        self.all_optimizers = []
-        self.training_batch_size = training_batch_size
-        self.validation_batch_size = validation_batch_size
+
+        # --- Model ---
         self.model_scope_name = model_scope_name
         self.auto_bake = False
         self.scalar_values = {}
         self.scalar_values_validation = {}
+        self.learning_rate = self.editable_float('Learning_Rate', learning_rate)
+        self.all_optimizers = []
+        # --- Data ---
+        self.training = tf.placeholder(tf.bool, (), 'training')
+        self.training_batch_size = training_batch_size
+        self.validation_batch_size = validation_batch_size
+        self._placeholder_struct = None  # Data-placeholders or Iterator placeholder
+        self._training_set = None
+        self._validation_set = None
+        self._pipeline = None
         self.set_data(None, None)
         assert stride is None or epoch_size is None
         self.epoch_size = epoch_size if epoch_size is not None else stride
@@ -129,11 +156,41 @@ class LearningApp(App):
         self.validation_step()
         return self
 
+    def set_learning_rate(self, learning_rate):
+        self.set_editable_value('Learning_Rate', learning_rate)
+
     def set_data(self, dict, train=None, val=None):
+        """
+Specify what data to use for training and validation.
+
+The content of `dict` determines the data pipeline that is used.
+  - 'placeholder' pipline: the static TensorFlow graph uses placeholders as input. `dict` maps from placeholders to file names or Stream instances. Placeholders and corresponding streams may be placed inside structs.
+  - 'dataset_handle' pipeline: Use TensorFlow data pipeline. `dict` contains 'iterator_handle' and related properties as returned by `build_graph_input(...)[1]`.
+
+Regardless of pipeline, the recommended way to obtain `dict` is through `build_graph_input(...)[1]`.
+
+        :param dict: pipeline-dependent dict
+        :type dict: dict
+        :param train: (optional) Dataset used for training
+        :type train: Dataset
+        :param val: (optional) Dataset used for validation
+        :type val: Dataset
+        """
+        assert isinstance(train, Dataset) or train is None
+        assert isinstance(val, Dataset) or val is None
         if train is not None or val is not None:
             assert dict is not None
+        if train is not None and val is not None:
+            self.value_view_training_data = False
         self._training_set = train
         self._validation_set = val
+        if dict is not None and 'iterator_handle' in dict:
+            self._init_tf_pipeline(**dict)
+        else:
+            self._init_numpy_iterators(dict)
+
+    def _init_numpy_iterators(self, dict):
+        self._pipeline = 'placeholder'
         self._placeholder_struct = []
         self._channel_struct = []
         if dict is not None:
@@ -145,17 +202,31 @@ class LearningApp(App):
         # Train
         if self._training_set is not None:
             self._train_reader = BatchReader(self._training_set, self._channel_struct)
-            self._train_iterator = self._train_reader.all_batches(batch_size=self.training_batch_size, loop=True)
+            self._train_iterator = self._train_reader.all_batches(batch_size=self.world.batch_size or self.training_batch_size, loop=True)
         else:
             self._train_reader = None
             self._train_iterator = None
         # Val
         if self._validation_set is not None:
-            self.value_view_training_data = False
             self._val_reader = BatchReader(self._validation_set, self._channel_struct)
         else:
             self._val_reader = None
 
+    def _init_tf_pipeline(self, iterator_handle, names, shapes, dtypes, frames):
+        self._placeholder_struct = iterator_handle
+        self._pipeline = 'dataset_handle'
+        if self._training_set is not None:
+            assert isinstance(self._training_set, TFDataset)
+            if self._training_set.name is None:
+                self._training_set.name = 'train'
+            self._training_set.setup(names, shapes, dtypes, batch_size=self.training_batch_size, frames=frames)
+            self._training_set.reset_iterator(self.session)
+        if self._validation_set is not None:
+            assert isinstance(self._validation_set, TFDataset)
+            if self._validation_set.name is None:
+                self._validation_set.name = 'validation'
+            self._validation_set.setup(names, shapes, dtypes, batch_size=self.validation_batch_size, frames=frames)
+
     def add_objective(self, loss, name='Loss', optimizer=None, reg=None, vars=None):
         assert len(loss.shape) <= 1, 'Loss function must be a scalar'
         if not optimizer:
@@ -178,8 +249,10 @@ class LearningApp(App):
         return node
 
     def step(self):
-        self.optimization_step(self.all_optimizers)
-        if self.steps % self.epoch_size == 0:
+        optimized = self.optimization_step(self.all_optimizers)
+        if not optimized:
+            self.steps -= 1
+        if self._pipeline == 'placeholder' and self.steps % self.epoch_size == 0:
             self.validation_step(create_checkpoint=True)
         return self
 
@@ -188,9 +261,20 @@ class LearningApp(App):
             optim_nodes = list(optim_nodes)
         except:
             optim_nodes = [optim_nodes]
-        batch = next(self._train_iterator) if self._train_iterator is not None else None
+        if self._pipeline == 'placeholder':
+            batch = next(self._train_iterator) if self._train_iterator is not None else None
+        elif self._pipeline == 'dataset_handle':
+            batch = self._training_set.iterator_handle
+        else:
+            raise NotImplementedError('Pipeline %s' % self._pipeline)
         feed_dict = self._feed_dict(batch, True)
-        scalar_values = self.session.run(optim_nodes + self.scalars, feed_dict, summary_key='train', merged_summary=self.merged_scalars, time=self.steps)[len(optim_nodes):]
+        try:
+            scalar_values = self.session.run(optim_nodes + self.scalars, feed_dict, summary_key='train', merged_summary=self.merged_scalars, time=self.steps)[len(optim_nodes):]
+        except tf.errors.OutOfRangeError as error:
+            if self._pipeline != 'dataset_handle':
+                raise error
+            self.on_training_set_end()
+            return False
         self.scalar_values = {name: value for name, value in zip(self.scalar_names, scalar_values)}
         if log_loss is None:
             log_loss = self.log_scalars
@@ -199,12 +283,23 @@ class LearningApp(App):
         assert isinstance(log_loss, bool)
         if log_loss:
             self.info('Optimization (%06d): ' % self.steps + ', '.join([self.scalar_names[i] + ': ' + str(scalar_values[i]) for i in range(len(self.scalars))]))
+        return True
+
+    def on_training_set_end(self):
+        self.validation_step(create_checkpoint=True)
+        self._training_set.reset_iterator(self.session)
 
     def validation_step(self, create_checkpoint=False):
-        if self._val_reader is None:
+        if self._validation_set is None:
             return
-        batch = self._val_reader[0:self.validation_batch_size]
+        if self._pipeline == 'placeholder':
+            batch = self._val_reader[0:self.validation_batch_size]
+        elif self._pipeline == 'dataset_handle':
+            batch = self._validation_set.get_reset_handle(self.session)
+        else:
+            raise NotImplementedError('Pipeline %s' % self._pipeline)
         feed_dict = self._feed_dict(batch, False)
+        # ToDo iterate over complete valiadtion set and average the results, e.g. with tf.contrib.metrics.streaming_mean - https://stackoverflow.com/questions/40788785/how-to-average-summaries-over-multiple-batches
         scalar_values = self.session.run(self.scalars, feed_dict, summary_key='val', merged_summary=self.merged_scalars, time=self.steps)
         self.scalar_values_validation = {name: value for name, value in zip(self.scalar_names, scalar_values)}
         if create_checkpoint:
@@ -215,6 +310,14 @@ class LearningApp(App):
         return {}
 
     def _feed_dict(self, batch, training):
+        """
+Assemble a complete feed dict for graph execution.
+        :param batch: (optional)
+          'placeholder' pipeline: struct of Numpy arrays matching self._placeholder_struct
+          'dataset_handle' pipeline: iterator_handle string value
+        :param training: bool
+        :return: dict that can be passed to session.run()
+        """
         feed_dict = self.base_feed_dict()
         feed_dict.update(self.editable_values_dict())
         feed_dict[self.training] = training
@@ -222,9 +325,6 @@ class LearningApp(App):
             feed_dict[self._placeholder_struct] = batch
         return feed_dict
 
-    # def val(self, fetches, subrange=None):
-    #     return self.session.run(fetches, self._feed_dict(self.val_iterator, False, subrange=subrange))
-
     @property
     def view_reader(self):
         if self._val_reader is None and self._train_reader is None:
@@ -233,12 +333,27 @@ class LearningApp(App):
             return self._train_reader
         return self._train_reader if self.value_view_training_data else self._val_reader
 
+    def _view_dataset(self):
+        if self._validation_set is None and self._training_set is None:
+            return None
+        if self._validation_set is None:
+            return self._training_set
+        if self.value_view_training_data:
+            return self._training_set
+        else:
+            return self._validation_set
+
     def view(self, tasks):
         if tasks is None:
             return None
-        reader = self.view_reader
-        batch = reader[0:self.validation_batch_size] if reader is not None else None
-        return self.session.run(tasks, self._feed_dict(batch, False))
+        if self._pipeline == 'placeholder':
+            batch = self.view_reader[0:self.validation_batch_size] if self.view_reader is not None else None
+        elif self._pipeline == 'dataset_handle':
+            batch = self._view_dataset().get_reset_handle(self.session)
+        else:
+            raise NotImplementedError('Pipeline %s' % self._pipeline)
+        feed_dict = self._feed_dict(batch, False)
+        return self.session.run(tasks, feed_dict)
 
     @property
     def viewed_batch(self):
@@ -254,6 +369,9 @@ class LearningApp(App):
         self.session.save(dir)
         return dir
 
+    def action_save_model(self):
+        self.save_model()
+
     def load_model(self, checkpoint_dir):
         self.session.restore(checkpoint_dir, scope=self.model_scope_name)
 
@@ -266,12 +384,20 @@ class LearningApp(App):
         :param name: channel name
         :param field: Tensor, string (database fieldname) or function
         """
-        if istensor(field):
+        if is_tensorflow_field(field):
             App.add_field(self, name, lambda: self.view(field))
-        # elif isinstance(field, StructAttributeGetter):
-        #     App.add_field(self, name, lambda: self.view_batch(field))
         else:
             App.add_field(self, name, field)
 
 
 TFApp = LearningApp
+
+
+def is_tensorflow_field(obj):
+    if TF_BACKEND.is_tensor(obj):
+        return True
+    if isinstance(obj, StaggeredGrid):
+        return np.any([is_tensorflow_field(grid) for grid in obj.data])
+    if isinstance(obj, Field):
+        return TF_BACKEND.is_tensor(obj.data)
+    return False
