@@ -1,17 +1,10 @@
 import json
-import logging
 import os
-import socket
 import threading
 
-import tensorflow as tf
+from . import tf
 from tensorflow.python.client import timeline
 
-if tf.__version__[0] == '2':
-    logging.info('Adjusting for tensorflow 2.0')
-    tf = tf.compat.v1
-    tf.disable_eager_execution()
-
 
 class Timeliner:
 
