@@ -2,30 +2,24 @@
 import logging
 import warnings
 import numpy as np
-import tensorflow as tf
 from tensorflow.python import pywrap_tensorflow
+from . import tf, TF_BACKEND
 
-from phi import struct
+from phi import struct, math
 from phi.math.math_util import is_static_shape
 from phi.physics.field.staggered_grid import StaggeredGrid
 from phi.physics.field.grid import CenteredGrid
 
-if tf.__version__[0] == '2':
-    logging.info('Adjusting for tensorflow 2.0')
-    tf = tf.compat.v1
-    tf.disable_eager_execution()
-
 
 def _tf_name(trace, basename):
     path = trace.path('/')
     if basename is None and len(path) == 0:
         return None
     result = path if basename is None else basename + '/' + path
-    print(result)
     return result
 
 
-def placeholder(shape, dtype=np.float32, basename='Placeholder'):
+def placeholder(shape, dtype=None, basename='Placeholder'):
     if struct.isstruct(dtype):
         def placeholder_map(trace):
             shape, dtype = trace.value
@@ -33,7 +27,7 @@ def placeholder(shape, dtype=np.float32, basename='Placeholder'):
         zipped = struct.zip([shape, dtype], leaf_condition=is_static_shape)
         return struct.map(placeholder_map, zipped, leaf_condition=is_static_shape, trace=True)
     else:
-        def f(trace): return tf.placeholder(dtype, trace.value, _tf_name(trace, basename))
+        def f(trace): return tf.placeholder(TF_BACKEND.precision_dtype if dtype is None else dtype, trace.value, _tf_name(trace, basename))
         return struct.map(f, shape, leaf_condition=is_static_shape, trace=True)
 
 
@@ -45,26 +39,75 @@ def placeholder_like(obj, basename='Placeholder'):
     return struct.map(f, obj, leaf_condition=is_static_shape, trace=True)
 
 
-def variable(initial_value, dtype=np.float32, basename='Variable', trainable=True):
-    def f(attr): return tf.Variable(attr.value, name=_tf_name(attr, basename), dtype=dtype, trainable=trainable)
+def variable(initial_value, dtype=None, basename='Variable', trainable=True):
+    def f(attr): return tf.Variable(attr.value, name=_tf_name(attr, basename), dtype=TF_BACKEND.precision_dtype if dtype is None else dtype, trainable=trainable)
     return struct.map(f, initial_value, trace=True)
 
 
-def variable_generator(initializer, dtype=np.float32, basename='Variable', trainable=True):
+def variable_generator(initializer, dtype=None, basename='Variable', trainable=True):
     def create_variable(shape):
         initial_value = initializer(shape)
         return variable(initial_value, dtype, basename, trainable)
     return create_variable
 
 
-def isplaceholder(obj):
+def constant(value, dtype=None, basename='const'):
+    def f(trace): return tf.constant(trace.value, dtype=TF_BACKEND.precision_dtype if dtype is None else dtype, name=_tf_name(trace, basename))
+    return struct.map(f, value, trace=True)
+
+
+def is_placeholder(obj):
     return isinstance(obj, tf.Tensor) and obj.op.type == 'Placeholder'
 
 
+isplaceholder = is_placeholder
+
+
+def dataset_handle(shape, dtype, frames=None):
+    """
+Creates a single virtual TensorFlow dataset (iterator_handle) for the given struct.
+The dataset is expected to hold contain all fields required for loading the obj given the current context item condition.
+From the dataset, graph input tensors are derived and arranged into a struct of the same shape as obj.
+If an integer is passed to frames, a list of such structs is created by unstacking the second-outer-most dimension of the dataset.
+    :param shape: tensor shape or struct of tensor shapes
+    :param dtype: data type of struct of data types matching shape
+    :param frames: Number of frames contained in each example of the dataset. Expects shape (batch_size, frames, ...)
+    :type frames: int or None
+    :return: list of struct and placeholder.
+     1. If frames=None: valid struct corresponding to obj. If frames>1: list thereof
+     2. placeholder for a TensorFlow dataset iterator handle (dtype=string)
+    :rtype: tuple
+    """
+    shapes = tuple(struct.flatten(shape, leaf_condition=is_static_shape))
+    if struct.isstruct(dtype):
+        dtypes = tuple(struct.flatten(dtype))
+        assert len(dtypes) == len(shapes)
+    else:
+        dtypes = [dtype] * len(shapes)
+    if frames is not None:
+        shapes = tuple([shape[0:1] + (frames,) + shape[1:] for shape in shapes])
+    # --- TF Dataset handle from string ---
+    iterator_handle = tf.placeholder(tf.string, shape=[], name='dataset_iterator_handle')
+    iterator = tf.data.Iterator.from_string_handle(iterator_handle, output_types=dtypes, output_shapes=shapes)
+    next_element = iterator.get_next()
+    # --- Create resulting struct by splitting `next_element`s ---
+    if frames is None:
+        next_element_list = list(next_element)
+        next_struct = struct.map(lambda _: next_element_list.pop(0), shape, leaf_condition=is_static_shape)
+    else:
+        # --- Remap structures -> to `frames` long list of structs ---
+        next_struct = []
+        for frame_idx in range(frames):
+            next_element_list = list(next_element)
+            frame_struct = struct.map(lambda _: next_element_list.pop(0)[:, frame_idx, ...], shape, leaf_condition=is_static_shape)
+            next_struct.append(frame_struct)
+    return next_struct, iterator_handle
+
+
 def group_normalization(x, group_count, eps=1e-5):
     batch_size, H, W, C = tf.shape(x)
-    gamma = tf.Variable(np.ones([1, 1, 1, C]), dtype=tf.float32, name="GN_gamma")
-    beta = tf.Variable(np.zeros([1, 1, 1, C]), dtype=tf.float32, name="GN_beta")
+    gamma = tf.Variable(np.ones([1, 1, 1, C]), dtype=TF_BACKEND.precision_dtype, name="GN_gamma")
+    beta = tf.Variable(np.zeros([1, 1, 1, C]), dtype=TF_BACKEND.precision_dtype, name="GN_beta")
     x = tf.reshape(x, [batch_size, group_count, H, W, C // group_count])
     mean, var = tf.nn.moments(x, [2, 3, 4], keep_dims=True)
     x = (x - mean) / tf.sqrt(var + eps)
@@ -146,6 +189,7 @@ def residual_block_1d(y, nb_channels, kernel_size=(3,), _strides=(1,), activatio
 
 
 def istensor(obj):
+    warnings.warn("istensor is deprecated, use phi.tf.app.is_tensorflow_field instead", DeprecationWarning)
     if isinstance(obj, CenteredGrid):
         return istensor(obj.data)
     if isinstance(obj, StaggeredGrid):
@@ -153,6 +197,28 @@ def istensor(obj):
     return isinstance(obj, (tf.Tensor, tf.Variable))
 
 
+def gradients(y, xs, grad_y=None):
+    """
+    Compute the analytic gradients using TensorFlow's automatic differentiation.
+
+    :param y: tensor or struct of tensors. The contributions of all tensors in `y` are added up.
+    :param xs: struct of input tensors
+    :return: struct compatible with `xs` holding dy/dx
+    """
+    ys = struct.flatten(y)
+    if grad_y is not None:
+        grad_y = struct.flatten(grad_y)
+        for i in range(len(grad_y)):
+            grad_y[i] = math.cast(grad_y[i], math.dtype(ys[i]))
+    xs_ = struct.flatten(xs)
+    grad = tf.gradients(ys, xs_, grad_ys=grad_y)
+    return struct.unflatten(grad, xs)
+
+
+def stop_gradient(x):
+    return struct.map(tf.stop_gradient, x)
+
+
 def conv_function(scope, constants_file=None):
     if constants_file is not None:
         reader = pywrap_tensorflow.NewCheckpointReader(constants_file)
